pub mod dm_type;
pub mod dvm_value;
pub mod lib;
pub mod mem;
pub mod object_type;
pub mod op_codes;
pub mod platform;
pub mod util;

use crate::vm::dvm_value::DvmValue;
use crate::vm::lib::{DmConstant, DmLib};
use crate::vm::mem::{get_field_value, DmAllocObject};
use crate::vm::object_type::DmFn;
use crate::vm::platform::init_platform_functions;
use op_codes::*;
use std::alloc::{alloc, Layout};
use std::collections::HashMap;
use std::rc::Rc;
use util::{get_32_le, get_64_le};

pub type PlatformFunction = fn(args: Vec<DvmValue>, &mut DmVirtualMachine) -> DvmValue;

enum CallFrame {
    PlatformCall(PlatformCallFrame),
    DeimosCall(DeimosCallFrame),
}

pub struct PlatformCallFrame {
    pub name: String,
    pub args: Vec<DvmValue>,
}

struct DeimosCallFrame {
    return_address: usize,
}

pub struct DmVirtualMachine {
    libs: Vec<DmLib>,
    functions: HashMap<String, Rc<DmFn>>,
    platform_functions: HashMap<String, PlatformFunction>,
    ip: usize,
    call_stack: Vec<CallFrame>,
    registers: Vec<DvmValue>,
    register_state_stack: Vec<Vec<DvmValue>>,
}

fn load_functions(destination: &mut HashMap<String, Rc<DmFn>>, libs: &Vec<DmLib>) {
    for module in libs {
        for lib_fn in &module.functions {
            destination.insert(lib_fn.fqn.clone(), lib_fn.clone());
        }
        for interface_function in module
            .interfaces
            .iter()
            .flat_map(|interface| interface.get_functions())
        {
            destination.insert(interface_function.fqn.clone(), interface_function.clone());
        }
        for implementation_function in module
            .implementations
            .iter()
            .flat_map(|implementation| &implementation.functions)
        {
            destination.insert(
                implementation_function.fqn.clone(),
                implementation_function.clone(),
            );
        }
    }
}

impl DmVirtualMachine {
    pub fn new(libs: Vec<DmLib>) -> DmVirtualMachine {
        let mut vm = DmVirtualMachine {
            libs: libs,
            functions: HashMap::new(),
            platform_functions: init_platform_functions(),
            ip: 0,
            registers: Vec::new(),
            call_stack: Vec::new(),
            register_state_stack: Vec::new(),
        };
        load_functions(&mut vm.functions, &vm.libs);
        vm
    }

    pub fn get_fn_by_fqn(&self, fqn: &str) -> Option<Rc<DmFn>> {
        self.functions.get(fqn).cloned()
    }

    pub fn call_fn(&mut self, dm_function: &DmFn, args: Vec<DvmValue>) -> Option<DvmValue> {
        // save current state
        self.call_stack.push(CallFrame::DeimosCall(DeimosCallFrame {
            return_address: self.ip,
        }));
        self.register_state_stack.push(self.registers.clone());

        // zero registers and make sure there are enough for dm_function
        self.registers.clear();
        if self.registers.len() < args.len() {
            self.registers.resize(args.len(), DvmValue::Unit);
        }

        // push args
        for i in 0..args.len() {
            self.registers.insert(i, args[i].clone());
        }

        // run the byte code
        self.run_raw(&dm_function.byte_code);

        // restore state
        self.registers = self.register_state_stack.pop().unwrap();
        if let CallFrame::DeimosCall(deimos_call_frame) = self.call_stack.pop().unwrap() {
            self.ip = deimos_call_frame.return_address;
        }

        // return result
        self.registers.get(0).map(|x| x.clone())
    }

    pub fn run_raw(&mut self, code: &Vec<u8>) {
        let mut i = 0;
        while i < code.len() {
            let op_code = code[i];
            println!("op_code: {:#04x}", op_code);
            match code[i] {
                MOV_INT => {
                    let target_register = code[i + 1] as usize;
                    let operand = get_32_le!(code, i, 2, u32);
                    self.registers
                        .insert(target_register, DvmValue::Int(operand as i32));
                    i += 6;
                }
                MOV_LONG => {
                    let target_register = code[i + 1] as usize;
                    let operand = get_64_le!(code, i, 2, u64);
                    self.registers
                        .insert(target_register, DvmValue::Long(operand as i64));
                    i += 10;
                }
                MOV_DOUBLE => {
                    unimplemented!();
                }
                MOV_REGISTER => {
                    let target_register = code[i + 1] as usize;
                    let source_register = code[i + 2] as usize;
                    let source_value = self.registers.get(source_register).unwrap();
                    self.registers.insert(target_register, source_value.clone());
                    i += 3;
                }
                ALLOC => {
                    i += 1;
                    let target_register = code[i] as usize;
                    i += 1;
                    let alloc_size = get_32_le!(code, i, 0, usize);
                    i += 4;
                    let raw_implementation_name_length = get_32_le!(code, i, 0, usize);
                    i += 4;
                    let raw_implementation_name_bytes =
                        code[i..(i + raw_implementation_name_length)].to_vec();
                    i += raw_implementation_name_length;
                    let implementation_name =
                        String::from_utf8(raw_implementation_name_bytes).unwrap();

                    let implementation = self
                        .libs
                        .iter()
                        .find_map(|lib| {
                            lib.implementations
                                .iter()
                                .find(|implementation| implementation.fqn == implementation_name)
                        })
                        .expect(&format!(
                            "Implementation not found: {}",
                            implementation_name
                        ));

                    let layout = Layout::from_size_align(alloc_size, 1).unwrap();
                    let raw_data_ptr;
                    unsafe {
                        raw_data_ptr = alloc(layout);
                    }
                    let dm_alloc_object = DmAllocObject {
                        data: raw_data_ptr,
                        layout,
                        size: alloc_size,
                        implementation: implementation.clone(),
                    };

                    let dvm_value = DvmValue::Pointer(Rc::new(dm_alloc_object));
                    self.registers.insert(target_register, dvm_value);
                }
                DEALLOC => {
                    todo!()
                }
                MOV_INT_TO => {
                    todo!()
                }
                MOV_REGISTER_TO => {
                    i += 1;
                    let target_register = code[i] as usize;
                    i += 1;
                    let offset = get_32_le!(code, i, 0, isize);
                    i += 4;
                    let source_register = code[i] as usize;
                    i += 1;

                    let target_value = self.registers.get(target_register).unwrap();
                    let target_pointer = match target_value {
                        DvmValue::Pointer(alloc_object) => alloc_object.data,
                        _ => panic!("Expected a DvmValue::Pointer, but got: {:?}", target_value),
                    };

                    let source_value = self.registers.get(source_register).unwrap();
                    match source_value {
                        DvmValue::Pointer(source_alloc_object) => unsafe {
                            let source_pointer = Rc::into_raw(source_alloc_object.clone());
                            for (j, b) in (source_pointer as usize).to_le_bytes().iter().enumerate()
                            {
                                target_pointer.offset(offset + j as isize).write(*b);
                            }
                        },
                        _ => panic!("Currently unsupported source value: {:?}", source_value),
                    }
                }
                MOV_CONST => {
                    i += 1;
                    let target_register = code[i] as usize;
                    i += 1;
                    let lib_name_length = get_32_le!(code, i, 0, usize);
                    i += 4;
                    let lib_name_raw = code[i..(i + lib_name_length)].to_vec();
                    i += lib_name_length;
                    let const_id = get_32_le!(code, i, 0, usize);
                    i += 4;

                    let lib_name = String::from_utf8(lib_name_raw).unwrap();

                    if let Some(lib) = self.libs.iter().find(|&lib| lib.name == lib_name) {
                        let constant = &lib.constants[const_id];
                        match constant {
                            DmConstant::String(s) => {
                                let alloc_fn = self.get_fn_by_fqn("std::unsafe::mem::alloc")
                                    .expect("Could not find std::unsafe::mem::alloc");
                                let alloc_size_arg = DvmValue::Long(s.len() as i64);
                                let alloc_return_value = self.call_fn(&alloc_fn, vec![alloc_size_arg]).unwrap();
                                let DvmValue::Pointer(pointer_object) = alloc_return_value else {
                                    panic!("Expected std::unsafe::mem::alloc to return DvmValue::Pointer, but got: {:?}", alloc_return_value);
                                };
                                let raw_address_field = pointer_object.implementation.get_field("raw_address", &pointer_object)
                                    .expect("Could not get PointerImpl.raw_address field.");
                                unsafe {
                                    let raw_address_value = get_field_value(&raw_address_field, &pointer_object);
                                    let DvmValue::Long(raw_address) = raw_address_value else {
                                        panic!("Expected PointerImpl.raw_address to be a DvmValue::Long, but got: {:?}", raw_address_value);
                                    };
                                    let ptr = raw_address as usize as *mut u8;
                                    for (j, b) in s.bytes().enumerate() {
                                        ptr.offset(j as isize).write(b);
                                    }
                                    self.registers.insert(target_register, DvmValue::Pointer(pointer_object));
                                }
                            }
                            _ => {
                                panic!("Invalid constant type");
                            }
                        }
                    } else {
                        panic!("Could not find lib: {}", lib_name);
                    }
                }
                PLATFORM_CALL => {
                    i += 1;

                    let symbol_name_length = get_32_le!(code, i, 0, usize);
                    i += 4;

                    let symbol_name_raw = code[i..(i + symbol_name_length)].to_vec();
                    i += symbol_name_length;

                    let symbol_name = String::from_utf8(symbol_name_raw).unwrap();

                    let return_register = code[i + 1] as usize;
                    i += 1;

                    let arg_registers_length = code[i] as usize;
                    i += 1;

                    let arg_registers = code[i..(i + arg_registers_length)].to_vec();
                    i += arg_registers_length;

                    let mut args = Vec::new();
                    for arg_register in arg_registers {
                        let register_value = self.registers.get(arg_register as usize).unwrap();
                        args.push(register_value.clone());
                    }

                    let platform_function_result = self.platform_functions.get(&symbol_name);
                    if platform_function_result.is_none() {
                        panic!("Unknown platform function {}", symbol_name);
                    }
                    let platform_function = platform_function_result.unwrap();

                    let call_result = platform_function(args, self);
                    self.registers.insert(return_register, call_result);
                }
                INVOKE_FN => {
                    unimplemented!()
                }
                INVOKE_VIRTUAL => {
                    unimplemented!();
                }
                INVOKE_METHOD => {
                    i += 1;
                    let name_length = get_32_le!(code, i, 0, usize);
                    i += 4;
                    let name_raw = code[i..(i + name_length)].to_vec();
                    i += name_length;
                    let arg_registers_length = code[i] as usize;
                    i += 1;
                    let arg_registers = code[i..(i + arg_registers_length)].to_vec();
                    i += arg_registers_length;

                    let method_name = String::from_utf8(name_raw).unwrap();

                    let method = self
                        .get_fn_by_fqn(&method_name)
                        .expect(&format!("Could not find method: {}", method_name));

                    let mut args: Vec<DvmValue> = Vec::new();
                    for arg_register in arg_registers {
                        let register_value = self.registers.get(arg_register as usize).unwrap();
                        args.push(register_value.clone());
                    }

                    self.call_fn(method.as_ref(), args);
                }
                INVOKE_DYNAMIC => {
                    unimplemented!()
                }
                RETURN => {
                    i += 1;
                }
                RETURN_REGISTER => {
                    unimplemented!()
                }
                op_code => panic!("Invalid or unimplemented op code: {:#04x}", op_code),
            }
        }
    }
}

#[cfg(test)]
mod dvm_run_tests {
    use super::*;

    macro_rules! assert_register {
        ( $expected: expr, $register: expr ) => {
            assert_eq!($expected, $register.unwrap().clone());
        };
    }

    #[test]
    fn mov_1_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 1);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run_raw(&code);
        assert_register!(DvmValue::Int(1), vm.registers.get(0));
    }

    #[test]
    fn move_65535_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0xffff);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run_raw(&code);
        assert_register!(DvmValue::Int(0xffff), vm.registers.get(0));
    }

    #[test]
    fn move_int_max_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0x0fff_ffff);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run_raw(&code);
        assert_register!(DvmValue::Int(0x0fff_ffff), vm.registers.get(0));
    }

    #[test]
    fn move_register() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 1, 1);
        add_mov_register(&mut code, 0, 1);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.registers.resize(2, DvmValue::Unit);
        vm.run_raw(&code);
        assert_register!(DvmValue::Int(1), vm.registers.get(0));
    }

    #[test]
    #[ignore]
    fn mov_int_to_register_as_address() {
        let mut code = Vec::new();
        add_alloc(&mut code, 0, 4);
        add_mov_int_to(&mut code, 0, 0, 0xff);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run_raw(&code);
        let actual = vm.registers.get(0).unwrap().clone();
        match actual {
            DvmValue::Pointer(value_box) => {
                assert_eq!(DvmValue::Int(0xff), *value_box);
            }
            _ => panic!("Target register is not a pointer"),
        }
    }

    #[test]
    #[ignore]
    fn alloc_and_dealloc_expect_register_cleared() {
        let mut code = Vec::new();
        add_alloc(&mut code, 0, 4);
        add_dealloc(&mut code, 0);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run_raw(&code);
        assert_register!(DvmValue::Unit, vm.registers.get(0));
    }
}
